/*
 * Copyright (c) 2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Sep 20 2016
 */
#define _XOPEN_SOURCE 600 /* linux does not have ptsname(3) otherwise */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>

#include "codels.h"


/* --- mrsim_open_pty ------------------------------------------------------ */

/* Open a pseudo terminal, create a symlink */
int
mrsim_open_pty(const char *device)
{
  int fd;

  /* open non-blocking */
  fd = open("/dev/ptmx", O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd < 0) return fd;

  /* create symlink */
  unlink(device);
  if (symlink(ptsname(fd), device)) return -1;

  /* configure */
  if (grantpt(fd) || unlockpt(fd)) return -1;

  return fd;
}


/* --- mrsim_close_pty ----------------------------------------------------- */

/* Close a pseudo terminal, remove symlink */
void
mrsim_close_pty(int fd, const char *device)
{
  unlink(device);
  close(fd);
}


/* --- mrsim_wait_msg ------------------------------------------------------ */

int
mrsim_wait_msg(int fd)
{
  struct pollfd pfds;
  int s;

  pfds.fd = fd;
  pfds.events = POLLIN;

  do {
    s = poll(&pfds, 1, 500/*ms*/);
  } while (s < 0 && errno == EINTR);

  if (pfds.revents & POLLHUP || s < 0) return -1;
  return s;
}


/* --- mrsim_recv_msg ------------------------------------------------------ */

/* returns: 0: timeout/incomplete, -1: error, 1: complete msg */

int
mrsim_recv_msg(int fd, struct mrsim_pty_s *pty)
{
  struct iovec iov[2];
  ssize_t s;
  uint8_t c;

  do {
    /* feed the ring  buffer */
    iov[0].iov_base = pty->buf + pty->w;
    iov[1].iov_base = pty->buf;

    if (pty->r > pty->w) {
      iov[0].iov_len = pty->r - pty->w - 1;
      iov[1].iov_len = 0;
    } else if (pty->r > 0) {
      iov[0].iov_len = sizeof(pty->buf) - pty->w;
      iov[1].iov_len = pty->r - 1;
    } else {
      iov[0].iov_len = sizeof(pty->buf) - pty->w - 1;
      iov[1].iov_len = 0;
    }

    if (iov[0].iov_len || iov[1].iov_len) {
      do {
        s = readv(fd, iov, 2);
      } while(s < 0 && errno == EINTR);

      if (s < 0) {
        if (pty->r == pty->w) return -1;
      } else if (s == 0) {
        if (pty->r == pty->w) return 0;
      } else
        pty->w = (pty->w + s) % sizeof(pty->buf);
    }

    while(pty->r != pty->w) {
      c = pty->buf[pty->r];
      pty->r = (pty->r + 1) % sizeof(pty->buf);

      switch(c) {
        case '^':
          pty->start = true;
          pty->escape = false;
          pty->len = 0;
          break;

        case '$':
          if (!pty->start) break;
          pty->start = false;
          return 1;

        case '!':
          pty->start = false;
          break;

        case '\\':
          pty->escape = true;
          break;

        default:
          if (!pty->start) break;
          if (pty->len >= sizeof(pty->msg)) {
            pty->start = false; break;
          }

          if (pty->escape) {
            c = ~c; pty->escape = false;
          }
          pty->msg[pty->len++] = c;
          break;
      }
    }
  } while(1);

  return 0;
}


/* --- mrsim_send_msg ------------------------------------------------------ */

static void	mrsim_encode(char x, char **buf);

int
mrsim_send_msg(int fd, const char *fmt, ...)
{
  va_list ap;
  ssize_t s;
  char buf[64], *w, *r;
  char c;

  if (fd < 0) return -1;

  w = buf;

  va_start(ap, fmt);
  *w++ = '^';
  while((c = *fmt++)) {
    while (w - buf > sizeof(buf)-8 /* 8 = worst case (4 bytes escaped) */) {
      do {
        s = write(fd, buf, w - buf);
      } while (s < 0 && errno == EINTR);
      if (s < 0) return -1;

      if (s > 0 && s < w - buf) memmove(buf, buf + s, w - buf - s);
      w -= s;
    }

    switch(c) {
      case '%': {
        switch(*fmt++) {
          case '1':
            mrsim_encode(va_arg(ap, int/*promotion*/), &w);
            break;

          case '2': {
            uint16_t x = va_arg(ap, int/*promotion*/);
            mrsim_encode((x >> 8) & 0xff, &w);
            mrsim_encode(x & 0xff, &w);
            break;
          }
          case '@': {
            uint16_t *x = va_arg(ap, uint16_t *);
            size_t l = va_arg(ap, size_t);
            while (l--) {
              mrsim_encode((*x >> 8) & 0xff, &w);
              mrsim_encode(*x & 0xff, &w);
              x++;
            }
            break;
          }

          case '4': {
            uint32_t x = va_arg(ap, uint32_t);
            mrsim_encode((x >> 24) & 0xff, &w);
            mrsim_encode((x >> 16) & 0xff, &w);
            mrsim_encode((x >> 8) & 0xff, &w);
            mrsim_encode(x & 0xff, &w);
            break;
          }
        }
        break;
      }

      default:
        mrsim_encode(c, &w);
    }
  }
  *w++ = '$';
  va_end(ap);

  r = buf;
  while (w > r) {
    do {
      s = write(fd, r, w - r);
    } while (s < 0 && errno == EINTR);
    if (s < 0) return -1;

    r += s;
  }

  return 0;
}

static void
mrsim_encode(char x, char **buf)
{
  switch (x) {
    case '^': case '$': case '\\': case '!':
      *(*buf)++ = '\\';
      x = ~x;
  }
  *(*buf)++ = x;
}
